from types import ModuleType
import json

import requests

from . import resources
from . import errors

# Create a dict of resource classes
RESOURCE_CLASSES = {}
for name, module in resources.__dict__.items():
    if isinstance(module, ModuleType) and name.capitalize() in module.__dict__:
        RESOURCE_CLASSES[name] = module.__dict__[name.capitalize()]

# Create a mapping of status codes to classes
STATUS_MAP = {}
for name, Klass in errors.__dict__.items():
    if isinstance(Klass, type) and issubclass(Klass, errors.RazorpayError):
        STATUS_MAP[Klass().status] = Klass


class Client:
    """Razorpay client class"""

    DEFAULTS = {
        'base_url': 'https://api.razorpay.com/v1'
    }

    CLIENT_OPTIONS = set(DEFAULTS.keys())
    QUERY_OPTIONS = set(['from', 'to', 'count', 'skip'])
    REQUEST_OPTIONS = set(['headers', 'params', 'data'])

    ALL_OPTIONS = CLIENT_OPTIONS | QUERY_OPTIONS | REQUEST_OPTIONS

    def __init__(self, session=None, auth=None, **options):
        """Initialize a Client object with session, optional auth handler, and options"""
        self.session = session or requests.Session()
        self.auth = auth
        # merge the provided options (if any) with the global DEFAULTS
        self.options = _merge(self.DEFAULTS, options)
        # intializes each resource, injecting this client object into the constructor
        for name, Klass in RESOURCE_CLASSES.items():
            setattr(self, name, Klass(self))

    def request(self, method, path, **options):
        """Dispatches a request to the Razorpay HTTP API"""
        options = self._merge_options(options)
        url = options['base_url'] + path
        request_options = self._parse_request_options(options)
        response = getattr(self.session, method)(url, auth=self.auth, **request_options)
        if response.status_code in STATUS_MAP:
            raise STATUS_MAP[response.status_code](response.json())
        elif response.status_code >= 500 and response.status_code < 600:
            raise errors.ServerError(response.json())
        else:
            return response.json()

    def get(self, path, **options):
        """Parses GET request options and dispatches a request."""
        query_options = self._parse_query_options(options)
        parameter_options = self._parse_parameter_options(options)
        query = _merge(query_options, parameter_options)  # options in the query takes precendence
        return self.request('get', path, params=query, **options)

    def post(self, path, data, **options):
        """Parses POST request options and dispatches a request."""
        parameter_options = self._parse_parameter_options(options)
        data = _merge(parameter_options, data)  # values in the data body takes precendence
        return self.request('post', path, data=data, **options)

    def _merge_options(self, *objects):
        """Merges one or more options objects with client's options and returns a new options object"""
        return _merge(self.options, *objects)

    def _parse_query_options(self, options):
        """Selects query string options out of the provided options object"""
        return self._select_options(options, self.QUERY_OPTIONS)

    def _parse_parameter_options(self, options):
        """Selects all unknown options (not query string, API, or request options)"""
        return self._select_options(options, self.ALL_OPTIONS, invert=True)

    def _parse_request_options(self, options):
        """Select and formats options to be passed to the 'requests' library's request methods"""
        request_options = self._select_options(options, self.REQUEST_OPTIONS)
        if 'params' in request_options:
            params = request_options['params']
            for key in params:
                if isinstance(params[key], bool):
                    params[key] = json.dumps(params[key])
        if 'data' in request_options:
            # remove empty 'options':
            if 'options' in request_options['data'] and len(request_options['data']['options']) == 0:
                del request_options['data']['options']
            # serialize 'data' to JSON, requests doesn't do this automatically:
            request_options['data'] = json.dumps(request_options['data'])
        return request_options

    def _select_options(self, options, keys, invert=False):
        """Selects the provided keys (or everything except the provided keys) out of an options object"""
        options = self._merge_options(options)
        result = {}
        for key in options:
            if (invert and key not in keys) or (not invert and key in keys):
                result[key] = options[key]
        return result


def _merge(*objects):
    """Merge one or more objects into a new object"""
    result = {}
    [result.update(obj) for obj in objects]
    return result



from .helpers import *


class TestClientPayment(ClientTestCase):

    @responses.activate
    def test_payment_all(self):
        result = mock_file('payment_collection')
        responses.add(GET, 'https://api.razorpay.com/v1/payments', status=200, body=json.dumps(result), match_querystring=True)
        self.assertEqual(self.client.payment.all(), result)

    @responses.activate
    def test_payment_all_with_options(self):
        result = mock_file('payment_collection_with_one_payment')
        responses.add(GET, 'https://api.razorpay.com/v1/payments?count=1', status=200, body=json.dumps(result), match_querystring=True)
        self.assertEqual(self.client.payment.all(count=1), result)

    @responses.activate
    def test_payment_fetch(self):
        result = mock_file('fake_payment')
        responses.add(GET, 'https://api.razorpay.com/v1/payments/' + self.payment_id, status=200, body=json.dumps(result), match_querystring=True)
        self.assertEqual(self.client.payment.fetch('fake_payment_id'), result)

    @responses.activate
    def test_payment_capture(self):
        result = mock_file('fake_captured_payment')
        responses.add(POST, 'https://api.razorpay.com/v1/payments/' + self.payment_id + '/capture', status=200, body=json.dumps(result), match_querystring=True)
        self.assertEqual(self.client.payment.capture(self.payment_id, amount=5100), result)



from .helpers import *


class TestClientRefund(ClientTestCase):

    @responses.activate
    def test_refund_all(self):
        result = mock_file('refund_collection')
        responses.add(GET, 'https://api.razorpay.com/v1/payments/' + self.payment_id + '/refunds', status=200, body=json.dumps(result), match_querystring=True)
        self.assertEqual(self.client.refund.all(self.payment_id), result)

    @responses.activate
    def test_refund_fetch(self):
        result = mock_file('fake_refund')
        responses.add(GET, 'https://api.razorpay.com/v1/payments/' + self.payment_id + '/refunds/' + self.refund_id, status=200, body=json.dumps(result), match_querystring=True)
        self.assertEqual(self.client.refund.fetch(self.payment_id, self.refund_id), result)

    @responses.activate
    def test_refund_create(self):
        result = mock_file('fake_refund')
        responses.add(POST, 'https://api.razorpay.com/v1/payments/' + self.payment_id + '/refund', status=200, body=json.dumps(result), match_querystring=True)
        self.assertEqual(self.client.refund.create(self.payment_id), result)

    @responses.activate
    def test_refund_create_partial(self):
        result = mock_file('fake_refund')
        responses.add(POST, 'https://api.razorpay.com/v1/payments/' + self.payment_id + '/refund', status=200, body=json.dumps(result), match_querystring=True)
        self.assertEqual(self.client.refund.create(self.payment_id, {'amount': 2000}), result)



class RazorpayError(Exception):
    """Base Razorpay error class"""
    def __init__(self, status=None, error=None):
        self.status = status
        self.error = error


class BadRequestError(RazorpayError):
    def __init__(self, error=None):
        super(BadRequestError, self).__init__(
            status=400,
            error=error
        )


class NoAuthorizationError(RazorpayError):
    def __init__(self, error=None):
        super(NoAuthorizationError, self).__init__(
            status=401,
            error=error
        )


class NotFoundError(RazorpayError):
    def __init__(self, error=None):
        super(NotFoundError, self).__init__(
            status=404,
            error=error
        )


class ServerError(RazorpayError):
    def __init__(self, error=None):
        status = 500
        if error:
            status = error.status
        super(ServerError, self).__init__(
            status=status,
            error=error
        )



from setuptools import setup

setup(
    name="razorpay",
    version="0.1.0",
    description="Razorpay Python Client",
    url="https://github.com/decached/razorpay",
    author="Akash Kothawale",
    author_email="akash@decached.com",
    license="MIT",
    install_requires=["requests"],
    packages=["razorpay"],
    keywords='razorpay payment gateway india',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ]
)



import json
import razorpay
import responses
import os
import unittest

from responses import GET, POST


def mock_file(filename):
    if filename == '':
        return ''
    file_path = os.path.join(os.path.dirname(__file__), 'mocks/' + filename + '.json')
    return open(file_path).read()


class ClientTestCase(unittest.TestCase):
    def setUp(self):
        self.payment_id = 'fake_payment_id'
        self.refund_id = 'fake_refund_id'
        self.client = razorpay.Client(auth=('fake_key_id', 'fake_key_secret'))
